/*
 * Copyright (C) 2010  Michael Morin
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the  v. 3.0  or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the Academic Free License ("AFL")
 * v. 3.0  for more details.
 *
 * You should have received a copy of the Academic Free License ("AFL") v. 3.0
 * along with this program.
 *
 */

/*
 * From the author:
 * This software uses the following third party libraries:
 *  - Choco Solver 2.1.3 (http://www.emn.fr/z-info/choco-solver)
 *  - Jung 2.0.1 (http://jung.sourceforge.net)
 * and is related to the following publication:
 *  M. Morin, A.P. Papillon, F. Laviolette, I. Abi-Zeid, and C.G. Quimper, 
 *  "Constraint Programming for Path Planning with Uncertainty: Solving the 
 *  Optimal Search Path problem," in Proceedings of the 18th Conference on 
 *  Principles and Practice of Constraint Programming, Qubec, Qc, Canada, 
 *  2012, pp. 988-1003.
 * 
 * Enhancements and comments are welcomed. Please let us know!  
 *
 * Best regards,
 *
 * Michael
 * http://www.michaelmorin.info
 */


/** 
 * @file OspProblem.java
 * @brief The Optimal Search Path (OSP) problem instance class.
 */

package ospCp2012;

import java.io.File;
import java.util.ArrayList;

import choco.Choco;
import choco.Options;
import choco.cp.model.CPModel;
import choco.cp.solver.CPSolver;
import choco.cp.solver.search.integer.branching.AssignVar;
import choco.cp.solver.search.integer.valiterator.DecreasingDomain;
import choco.cp.solver.search.integer.varselector.StaticVarOrder;
import choco.kernel.common.logging.ChocoLogging;
import choco.kernel.common.logging.Verbosity;
import choco.kernel.model.Model;
import choco.kernel.model.variables.integer.IntegerVariable;
import choco.kernel.solver.Solver;
import edu.uci.ics.jung.graph.Graph;


/**
 * @class OspProblem
 * @version 1.0
 * @author Michael Morin
 * @date Created on: 2011-03-25
 * @date Modified on: 2012-04-07
 *  
 * @brief  class OspProblem
 *
 * This class implements an Optimal Search Path (OSP) problem instance 
 * container. The 'solve' method is used to generate the CP model and to
 * solve it. 
 * 
 */
public class OspProblem extends Problem {
	
	Model mModel;
	Solver mSolver;
	
	// Instance
	InstanceConfig mInstanceConfig;
	OspConfig mConfig;
	int mVerboseLevel;
	
	// Environment
	Graph<Integer,Integer> mEnv; // The accessibility graph
	
	// Data
	double[] mPoc;
	int mSuInitLocation;
	int mN; // mEnv regions (vertices) count
	int mQ; // Effort amount (needed for a further extension to inter-region 
		// visibility, i.e., the Optimal Search Path problem with Visibility 
		// (OSPV problem))
	int mT; // Maximum path length, i.e., total number of time steps that are
		// allowed to search the accessibility graph
	
	// Detection model
	double mCurrentRegionPod;
	
	// Motion model
	double[][] mMotionModelD;
	
	// Solution
	OspSolution mSearchPlan;
	
	public OspProblem() {
		mInstanceConfig = null;
		mConfig = null;
		mEnv = null;
		mPoc = null;
		mN = 0;
		mQ = 1;
		mT = 1;
		mMotionModelD = null;
		mCurrentRegionPod = 0.8;
		mSearchPlan = null;
		mVerboseLevel = 0;
	}
	
	/**
	 * Generate an Optimal Search Path (OSP) problem instance.
	 * 
	 * @param instance an OSP problem instance configuration container.
	 */
	public void generateOsp(InstanceConfig instance) {
		super.mInstanceId = instance.getId();
		mInstanceConfig = instance;
		mConfig = null;
		mEnv = instance.getEnv();
		mN = mEnv.getVertexCount();
		mSuInitLocation = instance.getSuPosition();
		mPoc = instance.createPocDistribution();
		mQ = instance.getQ();
		mT = instance.getT();
		mCurrentRegionPod = instance.getCurrentRegionPod();
		mMotionModelD = instance.createMotionModel();
		mSearchPlan = new OspSolution(mN, mT);
		mMotionModelType = 1;
	}
	
	public double getCurrentRegionPod() {
		return mCurrentRegionPod;
	}
	
	public double[] getInitPoc() {
		return mPoc;
	}
	
	public double[][] getMotionModel() {
		return mMotionModelD;
	}
	
	public int getSuInitLocation() {
		return mSuInitLocation;
	}
	
	public int getN() {
		return mN;
	}
	
	public int getQ() {
		return mQ;
	}
	
	public int getT() {
		return mT;
	}
	
	/**
	 * Generate and solve an Optimal Search Path (OSP) CP model using the 
	 * default configuration.
	 */
	public boolean solve() {
		return solveUsingOspModel(new OspConfig(), true, super.mOutputCsvFile, 0, "");	
	}
	
	/**
	 * Generate and solve an Optimal Search Path (OSP) CP model using a 
	 * specific configuration.
	 * 
	 * @param config the configuration container
	 * @param outputToFile generate a file output or not
	 * @param outputFileName output file path
	 * @param verboseLevel verbose lever
	 * @param ospvChocoLoggingOutputFileName log file for choco solver
	 * @return true if everything is fine
	 */
	public boolean solveUsingOspModel(OspConfig config, boolean outputToFile, String outputFileName, int verboseLevel, String ospvChocoLoggingOutputFileName) {	
		// The environment has not been generated.
		if (mEnv == null) return false;	
		
		// Keep the config id for logging purpose
		super.mConfigId = config.getId();
		mConfig = config;
		
		mVerboseLevel = verboseLevel;
		
		// Compute the motion model relative poc used by some optimization
		// enhancing configurations
		double[][] poc_motion =  computeMotionModelRelativePoc();
		
		// Initialize the default probabilities upper and lower bounds
		double poc_lb = 0.0;
		double poc_ub = 1.0;
		double pos_lb = 0.0;
		double pos_ub = 1.0;
		double cos_lb = 0.0;
		double cos_ub = 1.0;
		
		if(outputToFile) {
			writeHeaderToCsvFile(outputFileName);
		}
		
		// *** Constants
		// None
			
		// *** Model
		Model m = new CPModel();
		
		
		// *** Variables
		// Note: All time-related indices are in {0, 1, ... T}.
		IntegerVariable[] var_su_loc_y = new IntegerVariable[mT + 1];
		IntegerVariable[][] var_poc = new IntegerVariable[mT][mN];	// These variables are not decision variables
		IntegerVariable[][] var_pos = new IntegerVariable[mT][mN];	// These variables are not decision variables
	
		var_su_loc_y[0] = Choco.makeIntVar("y(0)", 0, mN - 1, Options.V_ENUM);//, Options.V_NO_DECISION);
		for(int t = 1; t < mT + 1; ++t) {
			var_su_loc_y[t] = Choco.makeIntVar("y(" + t + ")", 0, mN - 1, Options.V_ENUM);
		}
		m.addVariables(var_su_loc_y);
		
		for(int s = 0; s < mN; ++s) {
			var_poc[0][s] = Choco.makeIntVar("poc(0," + s + ")", convertProbaToInt(mPoc[s], config.getProbaScaleupFactor()), 
					convertProbaToInt(mPoc[s], config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_NO_DECISION);
		}
		m.addVariables(var_poc[0]);
		
		for(int t = 1; t < mT; ++t) {
			for(int s = 0; s < mN; ++s) {
				if(config.UsePocUpperBoundFilter() == true) {
					var_poc[t][s] = Choco.makeIntVar("poc(" + t + "," + s + ")", convertProbaToInt(poc_lb, config.getProbaScaleupFactor()), 
							convertProbaToInt(poc_motion[t][s], config.getProbaScaleupFactor()) + 1, Options.V_BOUND, Options.V_NO_DECISION);
				} else {
					var_poc[t][s] = Choco.makeIntVar("poc(" + t + "," + s + ")", convertProbaToInt(poc_lb, config.getProbaScaleupFactor()), 
							convertProbaToInt(poc_ub, config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_NO_DECISION);
				}
			}
			m.addVariables(var_poc[t]);
		}
		
		// t \in {1, ..., T}
		for(int t = 0; t < mT; ++t) {
			var_su_loc_y[t] = Choco.makeIntVar("y(" + t + ")", 0, mN - 1, Options.V_ENUM);
			for(int s = 0; s < mN; ++s) {
				if(config.UsePosUpperBoundFilter() == true) {
					var_pos[t][s] = Choco.makeIntVar("pos(" + t + "," + s + ")", convertProbaToInt(pos_lb, config.getProbaScaleupFactor()), 
							convertProbaToInt(poc_motion[t][s] * mCurrentRegionPod, config.getProbaScaleupFactor()) + 1, Options.V_BOUND, Options.V_NO_DECISION);
				} else {
					var_pos[t][s] = Choco.makeIntVar("pos(" + t + "," + s + ")", convertProbaToInt(pos_lb, config.getProbaScaleupFactor()), 
							convertProbaToInt(pos_ub, config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_NO_DECISION);
				}
			}
			m.addVariables(var_pos[t]);			
		}
		m.addVariables(var_su_loc_y);
		
		// *** Objective function
		IntegerVariable[] obj_cos = new IntegerVariable[mT];
		if(mConfig.getModel() == OspConfig.MODEL_SUM) {
			if(verboseLevel > 1) {
				System.out.println("MODEL_SUM");
			}
			
			for(int t = 0; t < mT - 1; ++t) {
				obj_cos[t] = Choco.makeIntVar("COS(" + t + ")", convertProbaToInt(cos_lb, config.getProbaScaleupFactor()), 
						convertProbaToInt(cos_ub, config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_NO_DECISION);
			}
			obj_cos[mT - 1] = Choco.makeIntVar("COS(" + (mT - 1) + ")", convertProbaToInt(cos_lb, config.getProbaScaleupFactor()), 
					convertProbaToInt(cos_ub, config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_OBJECTIVE);
			m.addVariables(obj_cos);

			m.addConstraint(Choco.eq(obj_cos[0], Choco.sum(var_pos[0])));
			for(int t = 1; t < mT; ++t) {
				m.addConstraint(Choco.eq(obj_cos[t], Choco.plus(obj_cos[t - 1], Choco.sum(var_pos[t]))));
			}
			
		} else if(mConfig.getModel() == OspConfig.MODEL_MAX) {
			if(verboseLevel > 1) {
				System.out.println("MODEL_MAX");
			}
			
			for(int t = 0; t < mT - 1; ++t) {
				obj_cos[t] = Choco.makeIntVar("COS(" + t + ")", convertProbaToInt(cos_lb, config.getProbaScaleupFactor()), 
						convertProbaToInt(cos_ub, config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_NO_DECISION);
			}
			obj_cos[mT - 1] = Choco.makeIntVar("COS(" + (mT - 1) + ")", convertProbaToInt(cos_lb, config.getProbaScaleupFactor()), 
					convertProbaToInt(cos_ub, config.getProbaScaleupFactor()), Options.V_BOUND, Options.V_OBJECTIVE);

			m.addConstraint(Choco.eq(obj_cos[0], Choco.max(var_pos[0])));
			for(int t = 1; t < mT; ++t) {
				m.addConstraint(Choco.eq(obj_cos[t], Choco.plus(obj_cos[t - 1], Choco.max(var_pos[t]))));
			}
		}
	
		// *** Constraints
		// Search plan path constraints
		ArrayList<int[]> accessibility_feas_pairs = computeFeasiblePaths();
		m.addConstraint(Choco.feasPairAC(var_su_loc_y[1], var_su_loc_y[0], computeFeasiblePathsFirstMove()));
		for(int t = 2; t < mT + 1; ++t) {	
			m.addConstraint(Choco.feasPairAC(var_su_loc_y[t], var_su_loc_y[t - 1], accessibility_feas_pairs));
		}
		
		// COS criterion constraints		
		for(int t = 0; t < mT; ++t) {
			for(int r = 0; r < mN; ++r) {
				m.addConstraint(Choco.implies(Choco.eq(var_su_loc_y[t + 1],r), Choco.eq(var_pos[t][r], Choco.div(Choco.mult(var_poc[t][r], convertProbaToInt(mCurrentRegionPod, config.getProbaScaleupFactor())), config.getProbaScaleupFactor()))));
				m.addConstraint(Choco.implies(Choco.neq(var_su_loc_y[t + 1],r), Choco.eq(var_pos[t][r], 0)));
			}
		}
		
		int[][] scaled_motion_model_D_inverse = new int[mN][mN];
		for(int s = 0; s < mN; ++s) {
			for(int r = 0; r < mN; ++r) {
				scaled_motion_model_D_inverse[r][s] = convertProbaToInt(mMotionModelD[s][r], config.getProbaScaleupFactor());
			}
		}
		for(int t = 1; t < mT; ++t) {
			for(int r = 0; r < mN; ++r) {
				m.addConstraint(Choco.eq(var_poc[t][r], Choco.div(Choco.minus(Choco.scalar(var_poc[t-1],scaled_motion_model_D_inverse[r]), Choco.scalar(var_pos[t-1],scaled_motion_model_D_inverse[r])), config.getProbaScaleupFactor())));
			}
		}
		
		// *** Output
		if(verboseLevel > 0) {
			displayModelStatsToScreen(m);
		}

		// *** Solver
		CPSolver solver = new CPSolver();
		solver.read(m);
		
		// Configuration
		if(config.getBranchingStrategy() == OspConfig.BRANCHING_STRATEGY_STATIC) {
			if(verboseLevel > 1) {
				System.out.println("BRANCHING_STRATEGY_STATIC");
			}
			// We don't need restarts if the branching strategy is a static one.
			solver.setRestartLimit(0);
		
			IntegerVariable[] ordered_vars = new IntegerVariable[mT];
			int k = 0;
			for(int t = 1; t < mT + 1; ++t) {
				ordered_vars[k] = var_su_loc_y[t];
				++k;
			}
		
			solver.addGoal(new AssignVar(new StaticVarOrder(solver, solver.getVar(ordered_vars)), new DecreasingDomain()));
		} else if(config.getBranchingStrategy() == OspConfig.BRANCHING_STRATEGY_STATIC_FAS_DYNAMIC_VALUE_ITERATOR) {
			if(verboseLevel > 1) {
				System.out.println("BRANCHING_STRATEGY_STATIC_FAS_DYNAMIC_VALUE_ITERATOR");
			}
			// We don't need restarts with this kind of branching strategy 
			// decision variables are ordered.
			solver.setRestartLimit(0);
		
			IntegerVariable[] ordered_vars = new IntegerVariable[mT];
			int k = 0;
			for(int t = 1; t < mT + 1; ++t) {
				ordered_vars[k] = var_su_loc_y[t];
				++k;
			}
			
			IntegerVariable[] all_poc = new IntegerVariable[mT * mN];
			k = 0;
			for(int t = 0; t < mT; ++t) {
				for(int r = 0; r < mN; ++r) {
					all_poc[k] = var_poc[t][r];
					k++;
				}
			}
			
			solver.addGoal(new AssignVar(new StaticVarOrder(solver, solver.getVar(ordered_vars)), 
					new DynamicTotalDetectionValIterator(solver.getVar(var_su_loc_y),solver.getVar(all_poc),mT,mN,mEnv,mMotionModelD,mCurrentRegionPod))); 
		}
		
		// Solve
		File file = null;
		
		if(verboseLevel > 3) {
			if(ospvChocoLoggingOutputFileName.isEmpty() == false) {
				file  = new File(ospvChocoLoggingOutputFileName);
				ChocoLogging.setFileHandler(file);
				
			}
			ChocoLogging.setVerbosity(Verbosity.SOLUTION);
		} else if(verboseLevel > 1) {
			ChocoLogging.setEveryXNodes(1000);
			ChocoLogging.setVerbosity(Verbosity.VERBOSE);
		}
		
		if(config.getBacktrackLimit() > 0) {
			solver.setBackTrackLimit(config.getBacktrackLimit());
		}
		
		if(config.getTimeLimitMs() > 0) {
			solver.setTimeLimit(config.getTimeLimitMs());
		}
		
		if((config.getBranchingStrategy() == OspConfig.BRANCHING_STRATEGY_STATIC) 
			|| (config.getBranchingStrategy() == OspConfig.BRANCHING_STRATEGY_STATIC_FAS_DYNAMIC_VALUE_ITERATOR)) {
			solver.maximize(false); // Without restarts
		} else {
			solver.maximize(true); // With restarts
		}
		
		// Compute solution
		computeSolutionOsp(solver, var_su_loc_y);
		
		// *** Output
		if(verboseLevel > 0) {
			System.out.println("Real COS = " + mSearchPlan.getObjectiveValue());
			displayResultToScreen(solver);
		}
		
		if(verboseLevel > 2) {
			System.out.println("Solution = ");
			String line_1 = "   t:";
			String line_2 = "y(t):";
			for(int t = 0; t < mT + 1; ++t) {
				line_1 += " " + t;
				line_2 += " " + solver.getVar(var_su_loc_y[t]).getVal();
			}
			System.out.println(line_1);
			System.out.println(line_2);
			
			for(int t = 0; t < mT; ++t) {
				System.out.print("pos_" + t + " =");
				for(int r = 0; r < mN; ++r) {
					System.out.print(" " + solver.getVar(var_pos[t][r]).getVal());
				}
				System.out.println();
			}
		}
		
		if(outputToFile) {
			super.writeResultToCsvFile(m,solver,mSearchPlan,outputFileName);
		}
		
		ChocoLogging.flushLogs();
		
		// Free the model and the solver
		solver.clear();
		m = null;
		solver = null;
		
		return true;
	}
	
	private void computeSolutionOsp(Solver solver, IntegerVariable[] varSuLocY) {

		for(int t = 0; t < mT + 1; ++t) {
			mSearchPlan.setSuPosition(t, solver.getVar(varSuLocY[t]).getVal());
		}

		for(int t = 1; t < mT + 1; ++t) { 
			mSearchPlan.setSuEffortAllocation(t, solver.getVar(varSuLocY[t]).getVal(), mQ);
		}
		mSearchPlan.evaluate(mPoc, mMotionModelD, mCurrentRegionPod);
	}
	
	private double[][] computeMotionModelRelativePoc() {
		// Compute the motion model relative poc.
		double[] prev_poc_vector = new double[mN];
		double[] curr_poc_vector = new double[mN];
		double[] swap_poc_vector = null;
		
		double[][]poc_motion = new double[mT][mN];
		for(int r = 0; r < poc_motion[0].length; ++r) {
			poc_motion[0][r] = mPoc[r];
			prev_poc_vector[r] = poc_motion[0][r];
		}
		
		for(int t = 1; t < poc_motion.length; ++t) {
			for(int r = 0; r < poc_motion[t].length; ++r) {
				curr_poc_vector[r] = 0.0;
				for(int s = 0; s < poc_motion[t].length; ++s) {
						curr_poc_vector[r] += mMotionModelD[s][r] * prev_poc_vector[s];
				}
				poc_motion[t][r] = curr_poc_vector[r];
			}
			swap_poc_vector = prev_poc_vector;
			prev_poc_vector = curr_poc_vector;
			curr_poc_vector = swap_poc_vector;
		}
		
		return poc_motion;
	}
	
	private ArrayList<int[]> computeFeasiblePathsFirstMove() {
		ArrayList<int[]> accessibilityFeasPairs = new ArrayList<int[]>();
		
		for(int r = 0; r < mN; ++r) {
			if((mEnv.findEdge(mSuInitLocation, r) != null)) {
				accessibilityFeasPairs.add(new int[]{r, mSuInitLocation}); // r is accessible from s
			}
		}

		return accessibilityFeasPairs;
	}
	
	private ArrayList<int[]> computeFeasiblePaths() {
		ArrayList<int[]> accessibilityFeasPairs = new ArrayList<int[]>();
		
		for(int s = 0; s < mN; ++s) {
			for(int r = 0; r < mN; ++r) {
				if((mEnv.findEdge(s, r) != null)) {
					accessibilityFeasPairs.add(new int[]{r, s}); // r is accessible from s
				}
			}
		}

		return accessibilityFeasPairs;
	}
	
}

//End of file
