/*
 * Copyright (C) 2010  Michael Morin
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the  v. 3.0  or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the Academic Free License ("AFL")
 * v. 3.0  for more details.
 *
 * You should have received a copy of the Academic Free License ("AFL") v. 3.0
 * along with this program.
 *
 */

/*
 * From the author:
 * This software uses the following third party libraries:
 *  - Choco Solver 2.1.3 (http://www.emn.fr/z-info/choco-solver)
 *  - Jung 2.0.1 (http://jung.sourceforge.net)
 * and is related to the following publication:
 *  M. Morin, A.P. Papillon, F. Laviolette, I. Abi-Zeid, and C.G. Quimper, 
 *  "Constraint Programming for Path Planning with Uncertainty: Solving the 
 *  Optimal Search Path problem," in Proceedings of the 18th Conference on 
 *  Principles and Practice of Constraint Programming, Qubec, Qc, Canada, 
 *  2012, pp. 988-1003.
 * 
 * Enhancements and comments are welcomed. Please let us know!  
 *
 * Best regards,
 *
 * Michael
 * http://www.michaelmorin.info
 */


/** 
 * @file Problem.java
 * @brief The Problem class.
 */

package ospCp2012;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;

import choco.kernel.model.Model;
import choco.kernel.model.variables.integer.IntegerVariable;
import choco.kernel.solver.Solver;

/**
 * @class Problem
 * @version 1.0
 * @author Michael Morin
 * @date Created on: 2011-10-19
 * @date Modified on: 2012-04-04
 *  
 * @brief  class Problem
 *
 * This class implements a basic problem instance container. Derived 
 * classes specialize it by adding models and solvers, and/or algorithms to
 * solve the problem as well as container to store the problem instance
 * data (e.g., the accessibility graph, the motion model of the object).
 * 
 */
public abstract class Problem {
	
	// The following ids are used in log files.
	int mInstanceId = 0; // The instance config id.
	int mConfigId = 0; // The problem config id.
	int mDrawingGridWidth; // Used to draw the graph on a grid for outputs
	int mDrawingGridHeight; // Used to draw the graph on a grid for outputs
	int mMotionModelType; // The ID of the motino model type.
		// 0 = Stationary
		// 1 = Uniform on neighbors
	
	String mOutputCsvFile = "output.txt";
	
	abstract boolean solve();
	
	protected int convertProbaToInt(double proba, int proba_scale_up_factor) {
		long tmp = (long) (proba_scale_up_factor * proba);
		assert (tmp >= 0);
		assert (tmp <= (long) Integer.MAX_VALUE);
		return (int) tmp;
	}
	
	protected void displayModelStatsToScreen(Model m) {
		System.out.println("Constant count = " + m.getNbConstantVars());
		System.out.println("Integer variables count = " + m.getNbIntVars());
		System.out.println("Real variables count = " + m.getNbRealVars());
		System.out.println("Constraints count = " + m.getNbConstraints());
	}
	
	protected void displayResultToScreen(Solver s) {
		System.out.println("Objective value = " + s.getObjectiveValue());
		System.out.println("Feasible solution found? = " + s.isFeasible());
		System.out.println("Number of feasible solutions encountered = " + s.getSolutionCount());
		System.out.println("Problem solved in " + s.getTimeCount() + " milliseconds.");
		System.out.println("Problem solved in " + s.getBackTrackCount() + " backtracks.");
	}
	
	protected void writeHeaderToCsvFile(String outFile) {
		// Note: there's no need to optimize this output since the current 
		// time related performance metric only count solver's time.
		try {
			 // Write the header only if the file doesn't exist
			 if((new File(outFile)).exists() == false) {
				 PrintStream out = new PrintStream(new FileOutputStream(outFile,true));
				 out.println("instanceId,configId,NbConstantVars,NbConstantVars,NbRealVars,NbConstraints,ms,restarts,backtracks,isFeasible(0/1),modelObjective,realObjective");
				 out.close();
			 }
			 	
		} catch (FileNotFoundException e) {
		  e.printStackTrace();
		}
	}
	
	protected void writeResultToCsvFile(Model m, Solver s, Solution sol, String outFile) {
		// Note: there's no need to optimize this output since the current 
		// time related performance metric only count solver's time.
		try {
			 PrintStream out = new PrintStream(new FileOutputStream(outFile,true));
			 out.print(mInstanceId + ",");
			 out.print(mConfigId + ",");
			 out.print(m.getNbConstantVars() + ",");
			 out.print(m.getNbIntVars() + ",");
			 out.print(m.getNbRealVars() + ",");
			 out.print(m.getNbConstraints() + ",");
			 out.print(s.getTimeCount() + ",");
			 out.print(s.getRestartCount() + ",");
			 out.print(s.getBackTrackCount() + ",");
			 if(s.isFeasible() == true) {
				 out.print("1,");
			 } else {
				 out.print("0,");
			 }
			 out.print(s.getObjectiveValue() + ",");
			 out.println(sol.getObjectiveValue());
		
			 out.close();
		} catch (FileNotFoundException e) {
		  e.printStackTrace();
		}
	}
	
	protected IntegerVariable[] subArray(IntegerVariable[] array, int l, int u) {
		IntegerVariable[] r = new IntegerVariable[u - l + 1];
		for(int i = 0; i <= u - l; ++i) {
			r[i] = array[l+i];
		}
		return r;
	}
	
	protected IntegerVariable[] subArray(IntegerVariable[] array, ArrayList<Integer> indices) {
		IntegerVariable[] r = new IntegerVariable[indices.size()];
		for(int i = 0; i < indices.size(); ++i) {
			r[i] = array[indices.get(i)];
		}
		return r;
	}
}

//End of file
