/*
 * Copyright (C) 2010  Michael Morin
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the  v. 3.0  or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the Academic Free License ("AFL")
 * v. 3.0  for more details.
 *
 * You should have received a copy of the Academic Free License ("AFL") v. 3.0
 * along with this program.
 *
 */

/*
 * From the author:
 * This software uses the following third party libraries:
 *  - Choco Solver 2.1.3 (http://www.emn.fr/z-info/choco-solver)
 *  - Jung 2.0.1 (http://jung.sourceforge.net)
 * and is related to the following publication:
 *  M. Morin, A.P. Papillon, F. Laviolette, I. Abi-Zeid, and C.G. Quimper, 
 *  "Constraint Programming for Path Planning with Uncertainty: Solving the 
 *  Optimal Search Path problem," in Proceedings of the 18th Conference on 
 *  Principles and Practice of Constraint Programming, Qubec, Qc, Canada, 
 *  2012, pp. 988-1003.
 * 
 * Enhancements and comments are welcomed. Please let us know!  
 *
 * Best regards,
 *
 * Michael
 * http://www.michaelmorin.info
 */


/** 
 * @file ToyGraphGeneratorjava
 * @brief The ToyGraphGenerator class file.
 */

package ospCp2012;

import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.SparseMultigraph;
import edu.uci.ics.jung.graph.util.EdgeType;

/**
 * @class ToyGraphGenerator
 * @version 1.0
 * @author Michael Morin
 * @date Created on: 2011-03-22
 * @date Modified on: 2012-02-16
 *  
 * @brief  class ToyGraphGenerator
 *
 * This class purpose is to generate simple graphs.
 * <p>
 * The first set of problems contains simples graphs. Each graph of this set 
 * set has a unique id from 0 to mToyGraphCount - 1.
 * <p>
 * The second set of problems contains graphs from fictitious or real
 * environments. They are accessible via specific generate methods. For  
 * instance, to access the Universit Laval campus map graph, we use the  
 * method generateLavalCampusMap().
 */
public class ToyGraphGenerator {
	int mToyGraphCount = 4; // The total number of simple graphs.

	public Graph<Integer,Integer> generate(int graphId) {
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		if(graphId == 0) {
			return generateGraph0(g);
		} else if(graphId == 1) {
			return generateGraph1(g);
		} else if(graphId == 2) {
			return generateGraph2(g);
		} else if(graphId == 3) {
			return generateGraph3(g);
		} 
		
		return null;
	}
	
	public Graph<Integer,Integer> generateCycle(int length) {
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		addReflexiveVertices(length,g);
		for(int i = 0; i < length; ++i) {
			addUndirectedEdge(i, (i + 1) % length, g);
		}
		
		return g;
	}
	
	public Graph<Integer,Integer> generateCcumCongressExample() {
		// This is an example for the Ccum Congress beginning on June 15, 2011
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int n = 12;
		
		addReflexiveVertices(n, g);
		addUndirectedEdge(0, 1, g);
		addUndirectedEdge(0, 4, g);
		addUndirectedEdge(0, 5, g);
		addUndirectedEdge(1, 2, g);
		addUndirectedEdge(1, 5, g);
		addUndirectedEdge(2, 3, g);
		addUndirectedEdge(2, 5, g);
		addUndirectedEdge(2, 6, g);
		addUndirectedEdge(3, 6, g);
		addUndirectedEdge(4, 5, g);
		addUndirectedEdge(5, 6, g);
		addUndirectedEdge(5, 8, g);
		addUndirectedEdge(5, 9, g);		
		addUndirectedEdge(6, 7, g);
		addUndirectedEdge(7, 11, g);
		addUndirectedEdge(8, 9, g);
		addUndirectedEdge(10, 11, g);
		
		
		return g;
	}
	
	public Graph<Integer,Integer> generateNonCopwinBuildingExample() {
		// A simple graph example taken from
		// M. Morin, "Multi-Criteria Path Planning with Terrain Visibility 
		// Constraints: The Optimal Searcher Path Problem with Visibility," 
		// Universit Laval, Qubec, QC, Canada, Master's thesis, 2010.
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int n = 13;
		
		addReflexiveVertices(n, g);
		addUndirectedEdge(0, 4, g);
		addUndirectedEdge(1, 4, g);
		addUndirectedEdge(1, 2, g);
		addUndirectedEdge(2, 4, g);
		addUndirectedEdge(3, 4, g);
		addUndirectedEdge(4, 5, g);
		addUndirectedEdge(4, 9, g);
		addUndirectedEdge(4, 11, g);
		addUndirectedEdge(4, 8, g);
		addUndirectedEdge(6, 8, g);
		addUndirectedEdge(7, 8, g);
		addUndirectedEdge(9, 10, g);
		addUndirectedEdge(9, 11, g);
		addUndirectedEdge(10, 12, g);
		addUndirectedEdge(11, 12, g);
				
		return g;
	}
	
	public Graph<Integer,Integer> generateCopwinBuildingExample() {
		// A simple graph example taken from
		// M. Morin, "Multi-Criteria Path Planning with Terrain Visibility 
		// Constraints: The Optimal Searcher Path Problem with Visibility," 
		// Universit Laval, Qubec, QC, Canada, Master's thesis, 2010.
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int n = 13;
		
		addReflexiveVertices(n, g);
		addUndirectedEdge(0, 1, g);
		addUndirectedEdge(0, 6, g);
		addUndirectedEdge(1, 6, g);
		addUndirectedEdge(2, 6, g);
		addUndirectedEdge(3, 6, g);
		addUndirectedEdge(4, 7, g);
		addUndirectedEdge(5, 6, g);
		addUndirectedEdge(6, 7, g);
		addUndirectedEdge(7, 8, g);
		addUndirectedEdge(6, 9, g);
		addUndirectedEdge(6, 10, g);
		addUndirectedEdge(6, 11, g);
		addUndirectedEdge(6, 12, g);
		addUndirectedEdge(9, 10, g);
		addUndirectedEdge(9, 12, g);
		addUndirectedEdge(10, 11, g);
		addUndirectedEdge(11, 12, g);
				
		return g;
	}
	
	public Graph<Integer,Integer> generateLavalCampusMap() {
		// The Universit Laval tunnels map:
		// A graph discretization of the Universit Laval campus where each
		// vertex is one of the building that is accessible through the underground 
		// tunnels or one of the eleven main tunnels.
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int n = 34;
		
		addReflexiveVertices(n, g);
		addUndirectedEdge(0, 24, g);
		addUndirectedEdge(1, 24, g);
		addUndirectedEdge(2, 24, g);
		addUndirectedEdge(2, 25, g);
		addUndirectedEdge(3, 25, g);
		addUndirectedEdge(4, 25, g);
		addUndirectedEdge(25, 27, g);
		addUndirectedEdge(5, 27, g);
		addUndirectedEdge(6, 27, g);
		addUndirectedEdge(7, 25, g);
		addUndirectedEdge(25, 26, g);
		addUndirectedEdge(9, 26, g);
		addUndirectedEdge(8, 9, g);
		addUndirectedEdge(10, 26, g);
		addUndirectedEdge(11, 26, g);
		addUndirectedEdge(26, 29, g);
		addUndirectedEdge(21, 26, g);
		addUndirectedEdge(12, 29, g);
		addUndirectedEdge(11, 29, g);
		addUndirectedEdge(28, 29, g);
		addUndirectedEdge(13, 28, g);
		addUndirectedEdge(15, 28, g);
		addUndirectedEdge(14, 15, g);
		addUndirectedEdge(15, 16, g);
		addUndirectedEdge(17, 29, g);
		addUndirectedEdge(12, 32, g);
		addUndirectedEdge(18, 32, g);
		addUndirectedEdge(30, 32, g);
		addUndirectedEdge(19, 30, g);
		addUndirectedEdge(19, 31, g);
		addUndirectedEdge(19, 33, g);
		addUndirectedEdge(20, 31, g);
		addUndirectedEdge(22, 33, g);
		addUndirectedEdge(23, 33, g);
		
		return g;
	}
	
	public Graph<Integer,Integer> generateLavalCampusMapMini() {
		// The Universit Laval tunnels map:
		// A graph discretization of the Universit Laval campus where each
		// vertex is one of the building that is accessible through underground 
		// tunnels or one of the eleven main tunnels. 
		// The graph generated by this function is isomorph to the one generated
		// by the function generateLavalCampusMap.
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int n = 34;
		
		addReflexiveVertices(n, g);
		addUndirectedEdge(0,1,g);
		addUndirectedEdge(2,9,g);
		addUndirectedEdge(3,4,g);
		addUndirectedEdge(4,5,g);
		addUndirectedEdge(5,6,g);
		addUndirectedEdge(5,13,g);
		addUndirectedEdge(7,1,g);
		addUndirectedEdge(1,8,g);
		addUndirectedEdge(10,11,g);
		addUndirectedEdge(4,11,g);
		addUndirectedEdge(11,12,g);
		addUndirectedEdge(14,15,g);
		addUndirectedEdge(8,15,g);
		addUndirectedEdge(21,15,g);
		addUndirectedEdge(15,16,g);
		addUndirectedEdge(9,16,g);
		addUndirectedEdge(10,16,g);
		addUndirectedEdge(11,16,g);
		addUndirectedEdge(16,17,g);
		addUndirectedEdge(11,20,g);
		addUndirectedEdge(20,19,g);
		addUndirectedEdge(18,19,g);
		addUndirectedEdge(22,23,g);
		addUndirectedEdge(15,23,g);
		addUndirectedEdge(15,24,g);
		addUndirectedEdge(16,25,g);
		addUndirectedEdge(19,27,g);
		addUndirectedEdge(26,27,g);
		addUndirectedEdge(23,28,g);
		addUndirectedEdge(29,30,g);
		addUndirectedEdge(30,31,g);
		addUndirectedEdge(26,30,g);
		addUndirectedEdge(26,32,g);
		addUndirectedEdge(32,33,g);
		
		return g;
	}
	
	private Graph<Integer,Integer> generateGraph0(Graph<Integer,Integer> g) {
		
		g.addVertex((Integer)0);
		g.addVertex((Integer)1);
		g.addVertex((Integer)2);

		g.addEdge((Integer) 0, 0, 0, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 1, 1, 1, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 2, 2, 2, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 3, 3, 3, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 4, 0, 1, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 5, 1, 2, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 6, 2, 3, EdgeType.UNDIRECTED);
		g.addEdge((Integer) 7, 1, 3, EdgeType.UNDIRECTED);
		
		return g;
	}
	
	private Graph<Integer,Integer> generateGraph1(Graph<Integer,Integer> g) {
		int n = 10;
		
		addReflexiveVertices(n, g);
		
		addUndirectedEdge(0, 1, g);
		addUndirectedEdge(1, 2, g);
		addUndirectedEdge(2, 3, g);
		addUndirectedEdge(1, 3, g);
		
		addPath(3, 9, g);
		
		return g;
	}
	
	private Graph<Integer,Integer> generateGraph2(Graph<Integer,Integer> g) {
		int n = 4;
		
		addReflexiveVertices(n, g);
		
		// A cycle
		addUndirectedEdge(0, 1, g);
		addUndirectedEdge(1, 2, g);
		addUndirectedEdge(2, 3, g);
		addUndirectedEdge(3, 0, g);
		
		return g;
	}
	
	private Graph<Integer,Integer> generateGraph3(Graph<Integer,Integer> g) {
		int n = 10;
		
		addReflexiveVertices(n, g);
		
		// A cycle
		addUndirectedEdge(0, 1, g);
		addUndirectedEdge(1, 2, g);
		addUndirectedEdge(2, 3, g);
		addUndirectedEdge(3, 0, g);
		
		addPath(3, 9, g);
		
		return g;
	}
	
	// Insert new graphs here using the followingt template:
	/*private Graph<Integer,Integer> generateGraph4(Graph<Integer,Integer> g) {
		int n = 10;
		
		addReflexiveVertices(n, g);
		
		// A cycle
		//addUndirectedEdge(0, 1, g);
		//addUndirectedEdge(1, 2, g);
		//addUndirectedEdge(2, 3, g);
		//addUndirectedEdge(3, 0, g);
		
		//addPath(3, 9, g);
		
		return g;
	}*/
	
	public Graph<Integer,Integer> generatePlusGrid(int gridHeight, int gridWidth) {
		// Generate a grid with plus adjacency constraints (North, South, West, East).
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int current_n = gridHeight * gridWidth;
		
		int k = 0;
		
		addReflexiveVertices(current_n, g);
		
		// Create the grid		
		for(int i = 0; i < gridHeight; ++i) {
			for(int j = 0; j < gridWidth; ++j) {
				if(i < gridHeight - 1) {
					addUndirectedEdge(k, k + gridWidth, g);
				}
				if(j < gridWidth - 1) {
					addUndirectedEdge(k, k + 1, g);
				}
				++k;
			}
		}
		
		return g;
	}
	
	public Graph<Integer,Integer> generateStarGrid(int gridHeight, int gridWidth) {
		// Generate a grid with star adjacency constraints (NE N NW W SW S SE E).
		Graph<Integer,Integer> g = new SparseMultigraph<Integer, Integer>();
		
		int current_n = gridHeight * gridWidth;
		
		int k = 0;
		
		addReflexiveVertices(current_n, g);
		
		// Create the grid		
		for(int i = 0; i < gridHeight; ++i) {
			for(int j = 0; j < gridWidth; ++j) {
				if(i < gridHeight - 1) {
					addUndirectedEdge(k, k + gridWidth, g);
				}
				if(j < gridWidth - 1) {
					addUndirectedEdge(k, k + 1, g);
				}
				if((i < gridHeight - 1) && (j < gridWidth - 1)) {
					addUndirectedEdge(k, k + 1 + gridWidth, g);
				}
				if((i < gridHeight - 1) && (j > 0)) {
					addUndirectedEdge(k, k - 1 + gridWidth, g);
				}
				++k;
			}
		}
		
		return g;
	}
	
	private Graph<Integer,Integer> addReflexiveVertices(int nVertices, Graph<Integer,Integer> graphG) {
		int current_n = graphG.getVertexCount();
		for(int i = current_n; i < current_n + nVertices; ++i) {
			graphG.addVertex(i);
			
			// Add a loop.
			graphG.addEdge((Integer) graphG.getEdgeCount(), i, i, EdgeType.UNDIRECTED);
		}
		return graphG;
	}
	
	private Graph<Integer,Integer> addUndirectedEdge(int v1, int v2, Graph<Integer,Integer> graphG) {
		if(graphG.isNeighbor(v1, v2) == false) {
			graphG.addEdge((Integer) graphG.getEdgeCount(), v1, v2, EdgeType.UNDIRECTED);
		}
		return graphG;
	}
	
	private Graph<Integer,Integer> addPath(int v1, int v2, Graph<Integer,Integer> graphG) {
		// If v1 or v2 are not vertices, the graph is not modified.
		if((v1 < 0) || (v1 >= graphG.getVertexCount())) return graphG;
		if((v2 < 0) || (v2 >= graphG.getVertexCount())) return graphG;
		
		// Create a path between v1 and v2.
		for(int i = v1; i < v2; ++i) {
			graphG.addEdge((Integer) graphG.getEdgeCount(), i, i+1);
		}
		
		return graphG;
	}
	
}

//End of file
